using .MPI

"""
    OptComms

Composite type holding the different communicators for MPI-based optimization.

# Fields:
- world_comm::MPI.Comm  # The top-level communicator between all processors. Rank 0 on this communicator is the global head.
- group_comm::MPI.Comm  # Communicator linking all the group heads.  Rank 0 on this communicator is the global head.
- compute_comm::MPI.Comm  # Communicator linking all the processors in a particular group.  Rank 0 on this communicator is the group head.
"""
struct OptComms
  world_comm::MPI.Comm
  group_comm::MPI.Comm
  compute_comm::MPI.Comm
  ranks::NamedTuple{(:world, :group, :compute), Tuple{Int, Int, Int}}
  #targetComm::MPI.Comm
  #targetGroup::Int
  #chunkLength::Int
  #displacement::Int
end

"""
    setup_comms(procs_per_group,world_comm::MPI.Comm)

Split `world_comm` into groups where the communicator splitting attempts to
adhere as close to `procs_per_group` as possible, returning an `OptComms` type.

See also: [`OptComms`](@ref)
"""
function setup_comms(procs_per_group::Int,
                     world_comm::MPI.Comm)
  global_size = MPI.Comm_size(world_comm)
  world_rank = MPI.Comm_rank(world_comm)
  group_div, group_rem = divrem(global_size-1, procs_per_group)
  n_worker_groups = group_rem == 0 ? group_div : group_rem >= div(group_div, 2) ? group_div + 1 : group_div
  @debug "group_div = $group_div, group_rem = $group_rem, n_worker_groups = $n_worker_groups"
  n_procs_group = collect(1:n_worker_groups) .* procs_per_group
  group_id = world_rank != 0 ? !isnothing(findfirst(world_rank .<= n_procs_group)) ? findfirst(world_rank .<= n_procs_group) : n_worker_groups : 0
  #println(world_rank,' ',group_id)
  #groupSize = nGroups > 1 ? group_id < nGroups-1 ? group_div : group_div + group_rem : global_size
  compute_comm = MPI.Comm_split(world_comm, group_id, world_rank)
  compute_rank = MPI.Comm_rank(compute_comm)
  group_comm = MPI.Comm_split(world_comm, compute_rank, world_rank)
  ranks = NamedTuple{(:world, :group, :compute), Tuple{Int, Int, Int}}((world_rank,group_id,compute_rank))
  #procsPerTarget = procs_per_group
  #targetComm = compute_comm
  return OptComms(world_comm, group_comm, compute_comm, ranks)
end


"""
    @add_target!(prob, method, target, weight, label)

Construct a target from a method signature provided in `method.`  The
method signature *must* have keyword arugment `geometry = geometry_key`,
where `geometry_key` is the identifier returned by `@add_derived_geometry`.

Example:
```julia-repl
julia> using MPI, VMEC, StellaratorOptimization

julia> prob = StellOpt.Problem{Float64}();

julia> vmec_wrapper = EquilibriumWrapper(input_dictionary);

julia> add_equilibrium_wrapper!(prob, vmec_wrapper);

julia> vmec_surface = @add_derived_geometry!(prob, VmecSurface(0.5, equilibrium = vmec_wrapper));

julia> @add_target!(prob, VMEC.quasisymmetry_deviation(1, 4, 16, 16, geometry = vmec_surface), 0.0, 1.0, "qs(0.5)");
```
"""
macro add_mpi_target!(prob, method, target, weight, label)
  mod = @__MODULE__

  #method_call = Expr(method.head, method.args...)
  #sym, index = find_last_module(method)
  #method_call.args[1] = insert_main(method, sym, index)
  method_expr = Expr(:quote, method)
  return esc(quote
        # Need to evaluate an argument in the calling scope
        # so the values can be passed to the calling function
        # The first argument will always be the function to be called
        method_call = Expr($method_expr.head, $method_expr.args...)
        calling_mod, index = $mod._find_calling_module(method_call)
        escaped_mod = Symbol(@__MODULE__)
        method_call = $mod._insert_escaped_module(method_call, escaped_mod, calling_mod, index)
        for i in 2:length($method_expr.args)
            if Meta.isexpr($method_expr.args[i], :kw)
                method_call.args[i].args[2] = eval($method_expr.args[i].args[2])
            else
                method_call.args[i] = eval($method_expr.args[i])
            end
        end
        geometry_slot = findfirst(x -> x isa UInt, method_call.args)
        geometry_key = !isnothing(geometry_slot) ? method_call.args[geometry_slot] : error("Target must have a UInt identifier for the geometry")
        comm_slot = findfirst(x -> x isa $mod.OptComms, method_call.args)
        if !isnothing(comm_slot)
            method_call.args[comm_slot] = :mpi_comm
        end
        $mod.add_target!($prob, $mod.Target($target, $weight, method_call, geometry_key, $label))
  end)
  
end
