
"""
    optimize_mpi(prob::Problem,comms::OptComms)

Run an optimization problem defined by `prob` using MPI with communicators
defined in `comms`.
"""
function optimize_mpi(prob::Problem{T},
                      comms::OptComms;
                     ) where {T}
    world_rank = comms.ranks.world
    if world_rank == 0
        # Define the callback function that requires knowledge of the problem
        function cb_function(x)
            update_history!(x, prob)
        end

        # Extract initial state vector of BitType data
        initial_state_vector = state_vector_values(prob)

        # Set the intial gradient of each target to zero
        for i in 1:length(prob.targets)
            prob.targets[i].gradient = zeros(T, length(prob.state_vector))
        end

        # Augment the arguments of the optimizer options with the callback
        option_args = Dict{Symbol, Any}(:callback => cb_function)
        for f in fieldnames(Optim.Options)
            val = getfield(prob.optimizer.options, f)
            if !isnothing(val)
                option_args[f] = val
            end
        end
        new_options = (; option_args...)

        log_io = open(prob.output_directory*"stellopt_$(prob.output_tag)_log.jl", "w+")
        logger = SimpleLogger(log_io, Logging.Info)
        residual = with_logger(logger) do
            Optim.optimize(Optim.only_fg!((F, G, θ) -> objective_gradient!(F, G, θ, prob, comms; logger_io = log_io)),
                            initial_state_vector, prob.optimizer.method, Optim.Options(; new_options...))
        end
        flush(log_io)
        close(log_io)
        write_history_file(prob)
        MPI.bcast(true, 0, comms.world_comm)
        return residual
    else
        worker_loop(prob, comms)
        return nothing
    end
end

function objective_gradient!(F,
                             G,
                             x::AbstractVector,
                             prob::Problem,
                             comms::OptComms;
                             logger_io::Union{Nothing, IOStream} = nothing,
                            )
    # If the loop is in this function there is still work left to do; bcast false
    MPI.bcast(false, 0, comms.world_comm)
    state_vector = MPI.bcast(x ,0, comms.world_comm)
    adjust_state_vector!(prob, state_vector)
    obj, grad = (Inf, nothing)
    if isnothing(F)
        MPI.bcast(false, 0, comms.world_comm)
    else
        MPI.bcast(true, 0, comms.world_comm)
        obj = compute_objective(prob, comms)
        MPI.bcast(obj, 0, comms.world_comm)
        adjust_objective!(prob, obj)
    end
    if isnothing(G)
        MPI.bcast(false, 0, comms.world_comm)
    else
        if !isnan(obj)
            MPI.bcast(true, 0, comms.world_comm)
            grad = gradient(prob, comms)
            copyto!(G, grad)
            adjust_gradient!(prob, grad)
        else
            MPI.bcast(false, 0, comms.world_comm)
            copyto!(G, fill(NaN, length(x)))
        end
    end
    !isnothing(logger_io) ? flush(logger_io) : nothing
    if !isnothing(F)
        return obj
    end
    nothing
end

"""
    worker_loop(prob::Problem,comms::OptComms)

Listens for optimization state vector and calculation termination from the
global head node.  Exits when a finished value is broadcast from the global
head node.
"""
function worker_loop(prob::Problem,
                     comms::OptComms;
                    )
    finished = false

    while !finished
        # Check whether head is broadcasting if optimization is finished
        finished = MPI.bcast(finished, 0, comms.world_comm)
        if !finished
            # Get updated state vector values from head node
            state_vector = MPI.bcast(nothing, 0, comms.world_comm)
            adjust_state_vector!(prob, state_vector)
            # Check whether objective function calculation is required
            objective_needed = MPI.bcast(nothing, 0, comms.world_comm)
            if objective_needed
                compute_objective(prob, comms)
                obj = MPI.bcast(nothing, 0, comms.world_comm)
                adjust_objective!(prob, obj)
            end
            # Check whether gradient calculation is required
            gradient_needed = MPI.bcast(nothing, 0, comms.world_comm)
            if gradient_needed
                gradient(prob, comms)
            end
        end
    end
end

function compute_objective(prob::Problem{T},
                           comms::OptComms;
                          ) where {T}
    n_worker_groups = MPI.Comm_size(comms.group_comm) - 1

    # Compute the initial equilibrium and objective functions, broadcast
    # the equilibrium data to all processors. At this point all ranks
    # have the equilibrium input data from the previous iteration, thus
    # can immediately call the evaluate function
    if comms.ranks.world == 0
        F = zero(T)
        success = MPI.bcast(nothing, 1, comms.world_comm)
        if success
            for i in keys(prob.equilibrium_wrappers)
                eq_data = MPI.bcast(nothing, 1, comms.world_comm)
                update_equilibrium!(prob.equilibrium_wrappers[i], eq_data)
                update_input!(prob.equilibrium_wrappers[i])
                eq_mod = lowercase(string(equilibrium_type(prob.equilibrium_wrappers[i])))
                iteration = isassigned(prob.history, 1) ? prob.history[end].iteration + 1 : 0
                write_input_file(prob.equilibrium_wrappers[i], "input_$(eq_mod).opt_$(iteration)")
            end
            target_values = MPI.bcast(nothing, 1, comms.world_comm)
            foreach(p -> prob.targets[p.first].current_value = p.second, target_values)
            F = MPI.bcast(nothing, 1, comms.world_comm)
        else
            F = convert(T, NaN)
        end
        return F
    elseif comms.ranks.group == 1
        valid_eqs = fill(true, length(prob.equilibrium_wrappers))
        j = 0
        # Do the calculation on the first worker group
        for i in keys(prob.equilibrium_wrappers)
            j += 1
            translate!(prob.equilibrium_wrappers[i], prob.state_vector)
            eq = compute_equilibrium(prob.equilibrium_wrappers[i], comms.compute_comm)
            if !isnothing(eq) && typeof(eq) <: equilibrium_type(prob.equilibrium_wrappers[i])
                update_equilibrium!(prob.equilibrium_wrappers[i], eq)
                update_input!(prob.equilibrium_wrappers[i])
            else
                valid_eqs[j] = false
            end
        end

        if all(valid_eqs)
            MPI.bcast(true, 1, comms.world_comm)
            for i in keys(prob.equilibrium_wrappers)
                MPI.bcast(prob.equilibrium_wrappers[i].eq, 1, comms.world_comm)
            end
            target_values = compute_all_targets(prob, comms.compute_comm)
            target_values = MPI.bcast(target_values, 1, comms.world_comm)
            foreach(p -> prob.targets[p.first].current_value = p.second, target_values)
            #for i in 1:length(prob.targets)
            #    MPI.bcast(prob.targets[i].current_value, 1, comms.world_comm)
            #end
            if comms.ranks.compute == 0
                F = least_squares_objective(prob)
                MPI.bcast(F, 1, comms.world_comm)
            else
                F = MPI.bcast(nothing, 1, comms.world_comm)
            end
        else
            MPI.bcast(false, 1, comms.world_comm)
        end
        return nothing
    else
        success = MPI.bcast(nothing, 1, comms.world_comm)
        if success
            for i in keys(prob.equilibrium_wrappers)
                eq = MPI.bcast(nothing, 1, comms.world_comm)
                update_equilibrium!(prob.equilibrium_wrappers[i], eq)
                update_input!(prob.equilibrium_wrappers[i])
            end
            compute_all_targets(prob, comms.compute_comm)
            target_values = MPI.bcast(nothing, 1, comms.world_comm)
            foreach(p -> prob.targets[p.first].current_value = p.second, target_values)
            F = MPI.bcast(nothing, 1, comms.world_comm)
        end
        return nothing
    end
end

